@@ -22,18 +22,21 @@ import (
 	"strings"
 	"sync/atomic"
 
+	"github.com/go-logr/logr"
 	. "github.com/onsi/ginkgo"
 	. "github.com/onsi/gomega"
-	"k8s.io/client-go/rest"
-	"sigs.k8s.io/controller-runtime/pkg/cache"
-
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/types"
+	"k8s.io/client-go/rest"
 	"k8s.io/client-go/util/workqueue"
+
+	"sigs.k8s.io/controller-runtime/pkg/cache"
+	"sigs.k8s.io/controller-runtime/pkg/client"
+	"sigs.k8s.io/controller-runtime/pkg/config/v1alpha1"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
 	"sigs.k8s.io/controller-runtime/pkg/event"
 	"sigs.k8s.io/controller-runtime/pkg/handler"
@@ -46,24 +49,41 @@ import (
 
 type typedNoop struct{}
 
-func (typedNoop) Reconcile(reconcile.Request) (reconcile.Result, error) {
+func (typedNoop) Reconcile(context.Context, reconcile.Request) (reconcile.Result, error) {
 	return reconcile.Result{}, nil
 }
 
-var _ = Describe("application", func() {
-	var stop chan struct{}
+type testLogger struct {
+	logr.Logger
+}
 
+func (l *testLogger) Init(logr.RuntimeInfo) {
+}
+
+func (l *testLogger) Enabled(int) bool {
+	return true
+}
+
+func (l *testLogger) Info(level int, msg string, keysAndValues ...interface{}) {
+}
+
+func (l *testLogger) WithValues(keysAndValues ...interface{}) logr.LogSink {
+	return l
+}
+
+func (l *testLogger) WithName(name string) logr.LogSink {
+	return l
+}
+
+var _ = Describe("application", func() {
 	BeforeEach(func() {
-		stop = make(chan struct{})
 		newController = controller.New
 	})
 
-	AfterEach(func() {
-		close(stop)
+	noop := reconcile.Func(func(context.Context, reconcile.Request) (reconcile.Result, error) {
+		return reconcile.Result{}, nil
 	})
 
-	noop := reconcile.Func(func(req reconcile.Request) (reconcile.Result, error) { return reconcile.Result{}, nil })
-
 	Describe("New", func() {
 		It("should return success if given valid objects", func() {
 			By("creating a controller manager")
@@ -78,6 +98,32 @@ var _ = Describe("application", func() {
 			Expect(instance).NotTo(BeNil())
 		})
 
+		It("should return error if given two apiType objects in For function", func() {
+			By("creating a controller manager")
+			m, err := manager.New(cfg, manager.Options{})
+			Expect(err).NotTo(HaveOccurred())
+
+			instance, err := ControllerManagedBy(m).
+				For(&appsv1.ReplicaSet{}).
+				For(&appsv1.Deployment{}).
+				Owns(&appsv1.ReplicaSet{}).
+				Build(noop)
+			Expect(err).To(MatchError(ContainSubstring("For(...) should only be called once, could not assign multiple objects for reconciliation")))
+			Expect(instance).To(BeNil())
+		})
+
+		It("should return an error if For function is not called", func() {
+			By("creating a controller manager")
+			m, err := manager.New(cfg, manager.Options{})
+			Expect(err).NotTo(HaveOccurred())
+
+			instance, err := ControllerManagedBy(m).
+				Owns(&appsv1.ReplicaSet{}).
+				Build(noop)
+			Expect(err).To(MatchError(ContainSubstring("must provide an object for reconciliation")))
+			Expect(instance).To(BeNil())
+		})
+
 		It("should return an error if there is no GVK for an object, and thus we can't default the controller name", func() {
 			By("creating a controller manager")
 			m, err := manager.New(cfg, manager.Options{})
@@ -138,6 +184,34 @@ var _ = Describe("application", func() {
 			Expect(instance).NotTo(BeNil())
 		})
 
+		It("should override max concurrent reconcilers during creation of controller, when using", func() {
+			const maxConcurrentReconciles = 10
+			newController = func(name string, mgr manager.Manager, options controller.Options) (
+				controller.Controller, error) {
+				if options.MaxConcurrentReconciles == maxConcurrentReconciles {
+					return controller.New(name, mgr, options)
+				}
+				return nil, fmt.Errorf("max concurrent reconcilers expected %d but found %d", maxConcurrentReconciles, options.MaxConcurrentReconciles)
+			}
+
+			By("creating a controller manager")
+			m, err := manager.New(cfg, manager.Options{
+				Controller: v1alpha1.ControllerConfigurationSpec{
+					GroupKindConcurrency: map[string]int{
+						"ReplicaSet.apps": maxConcurrentReconciles,
+					},
+				},
+			})
+			Expect(err).NotTo(HaveOccurred())
+
+			instance, err := ControllerManagedBy(m).
+				For(&appsv1.ReplicaSet{}).
+				Owns(&appsv1.ReplicaSet{}).
+				Build(noop)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(instance).NotTo(BeNil())
+		})
+
 		It("should override rate limiter during creation of controller", func() {
 			rateLimiter := workqueue.DefaultItemBasedRateLimiter()
 			newController = func(name string, mgr manager.Manager, options controller.Options) (controller.Controller, error) {
@@ -160,6 +234,29 @@ var _ = Describe("application", func() {
 			Expect(instance).NotTo(BeNil())
 		})
 
+		It("should override logger during creation of controller", func() {
+
+			logger := &testLogger{}
+			newController = func(name string, mgr manager.Manager, options controller.Options) (controller.Controller, error) {
+				if options.Log.GetSink() == logger {
+					return controller.New(name, mgr, options)
+				}
+				return nil, fmt.Errorf("logger expected %T but found %T", logger, options.Log)
+			}
+
+			By("creating a controller manager")
+			m, err := manager.New(cfg, manager.Options{})
+			Expect(err).NotTo(HaveOccurred())
+
+			instance, err := ControllerManagedBy(m).
+				For(&appsv1.ReplicaSet{}).
+				Owns(&appsv1.ReplicaSet{}).
+				WithLogger(logr.New(logger)).
+				Build(noop)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(instance).NotTo(BeNil())
+		})
+
 		It("should prefer reconciler from options during creation of controller", func() {
 			newController = func(name string, mgr manager.Manager, options controller.Options) (controller.Controller, error) {
 				if options.Reconciler != (typedNoop{}) {
@@ -211,18 +308,20 @@ var _ = Describe("application", func() {
 	})
 
 	Describe("Start with ControllerManagedBy", func() {
-		It("should Reconcile Owns objects", func(done Done) {
+		It("should Reconcile Owns objects", func() {
 			m, err := manager.New(cfg, manager.Options{})
 			Expect(err).NotTo(HaveOccurred())
 
 			bldr := ControllerManagedBy(m).
 				For(&appsv1.Deployment{}).
 				Owns(&appsv1.ReplicaSet{})
-			doReconcileTest("3", stop, bldr, m, false)
-			close(done)
+
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+			doReconcileTest(ctx, "3", m, false, bldr)
 		}, 10)
 
-		It("should Reconcile Watches objects", func(done Done) {
+		It("should Reconcile Watches objects", func() {
 			m, err := manager.New(cfg, manager.Options{})
 			Expect(err).NotTo(HaveOccurred())
 
@@ -231,13 +330,15 @@ var _ = Describe("application", func() {
 				Watches( // Equivalent of Owns
 					&source.Kind{Type: &appsv1.ReplicaSet{}},
 					&handler.EnqueueRequestForOwner{OwnerType: &appsv1.Deployment{}, IsController: true})
-			doReconcileTest("4", stop, bldr, m, true)
-			close(done)
+
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+			doReconcileTest(ctx, "4", m, true, bldr)
 		}, 10)
 	})
 
 	Describe("Set custom predicates", func() {
-		It("should execute registered predicates only for assigned kind", func(done Done) {
+		It("should execute registered predicates only for assigned kind", func() {
 			m, err := manager.New(cfg, manager.Options{})
 			Expect(err).NotTo(HaveOccurred())
 
@@ -251,7 +352,7 @@ var _ = Describe("application", func() {
 				CreateFunc: func(e event.CreateEvent) bool {
 					defer GinkgoRecover()
 					// check that it was called only for deployment
-					Expect(e.Meta).To(BeAssignableToTypeOf(&appsv1.Deployment{}))
+					Expect(e.Object).To(BeAssignableToTypeOf(&appsv1.Deployment{}))
 					deployPrctExecuted = true
 					return true
 				},
@@ -261,7 +362,7 @@ var _ = Describe("application", func() {
 				CreateFunc: func(e event.CreateEvent) bool {
 					defer GinkgoRecover()
 					// check that it was called only for replicaset
-					Expect(e.Meta).To(BeAssignableToTypeOf(&appsv1.ReplicaSet{}))
+					Expect(e.Object).To(BeAssignableToTypeOf(&appsv1.ReplicaSet{}))
 					replicaSetPrctExecuted = true
 					return true
 				},
@@ -270,8 +371,8 @@ var _ = Describe("application", func() {
 			allPrct := predicate.Funcs{
 				CreateFunc: func(e event.CreateEvent) bool {
 					defer GinkgoRecover()
-					//check that it was called for all registered kinds
-					Expect(e.Meta).Should(Or(
+					// check that it was called for all registered kinds
+					Expect(e.Object).Should(Or(
 						BeAssignableToTypeOf(&appsv1.Deployment{}),
 						BeAssignableToTypeOf(&appsv1.ReplicaSet{}),
 					))
@@ -286,13 +387,13 @@ var _ = Describe("application", func() {
 				Owns(&appsv1.ReplicaSet{}, WithPredicates(replicaSetPrct)).
 				WithEventFilter(allPrct)
 
-			doReconcileTest("5", stop, bldr, m, true)
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+			doReconcileTest(ctx, "5", m, true, bldr)
 
 			Expect(deployPrctExecuted).To(BeTrue(), "Deploy predicated should be called at least once")
 			Expect(replicaSetPrctExecuted).To(BeTrue(), "ReplicaSet predicated should be called at least once")
 			Expect(allPrctExecuted).To(BeNumerically(">=", 2), "Global Predicated should be called at least twice")
-
-			close(done)
 		})
 	})
 
@@ -306,6 +407,16 @@ var _ = Describe("application", func() {
 			Expect(err).NotTo(HaveOccurred())
 		})
 
+		It("should support multiple controllers watching the same metadata kind", func() {
+			bldr1 := ControllerManagedBy(mgr).For(&appsv1.Deployment{}, OnlyMetadata)
+			bldr2 := ControllerManagedBy(mgr).For(&appsv1.Deployment{}, OnlyMetadata)
+
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+
+			doReconcileTest(ctx, "6", mgr, true, bldr1, bldr2)
+		})
+
 		It("should support watching For, Owns, and Watch as metadata", func() {
 			statefulSetMaps := make(chan *metav1.PartialObjectMetadata)
 
@@ -313,16 +424,25 @@ var _ = Describe("application", func() {
 				For(&appsv1.Deployment{}, OnlyMetadata).
 				Owns(&appsv1.ReplicaSet{}, OnlyMetadata).
 				Watches(&source.Kind{Type: &appsv1.StatefulSet{}},
-					&handler.EnqueueRequestsFromMapFunc{
-						ToRequests: handler.ToRequestsFunc(func(o handler.MapObject) []reconcile.Request {
-							ometa := o.Object.(*metav1.PartialObjectMetadata)
-							statefulSetMaps <- ometa
-							return nil
-						}),
-					},
+					handler.EnqueueRequestsFromMapFunc(func(o client.Object) []reconcile.Request {
+						defer GinkgoRecover()
+
+						ometa := o.(*metav1.PartialObjectMetadata)
+						statefulSetMaps <- ometa
+
+						// Validate that the GVK is not empty when dealing with PartialObjectMetadata objects.
+						Expect(o.GetObjectKind().GroupVersionKind()).To(Equal(schema.GroupVersionKind{
+							Group:   "apps",
+							Version: "v1",
+							Kind:    "StatefulSet",
+						}))
+						return nil
+					}),
 					OnlyMetadata)
 
-			doReconcileTest("8", stop, bldr, mgr, true)
+			ctx, cancel := context.WithCancel(context.Background())
+			defer cancel()
+			doReconcileTest(ctx, "8", mgr, true, bldr)
 
 			By("Creating a new stateful set")
 			set := &appsv1.StatefulSet{
@@ -383,7 +503,7 @@ type nonTypedOnlyCache struct {
 	cache.Cache
 }
 
-func (c *nonTypedOnlyCache) GetInformer(ctx context.Context, obj runtime.Object) (cache.Informer, error) {
+func (c *nonTypedOnlyCache) GetInformer(ctx context.Context, obj client.Object) (cache.Informer, error) {
 	switch obj.(type) {
 	case (*metav1.PartialObjectMetadata):
 		return c.Cache.GetInformer(ctx, obj)
@@ -396,14 +516,14 @@ func (c *nonTypedOnlyCache) GetInformerForKind(ctx context.Context, gvk schema.G
 }
 
 // TODO(directxman12): this function has too many arguments, and the whole
-// "nameSuffix" think is a bit of a hack It should be cleaned up significantly by someone with a bit of time
-func doReconcileTest(nameSuffix string, stop chan struct{}, blder *Builder, mgr manager.Manager, complete bool) {
+// "nameSuffix" think is a bit of a hack It should be cleaned up significantly by someone with a bit of time.
+func doReconcileTest(ctx context.Context, nameSuffix string, mgr manager.Manager, complete bool, blders ...*Builder) {
 	deployName := "deploy-name-" + nameSuffix
 	rsName := "rs-name-" + nameSuffix
 
 	By("Creating the application")
 	ch := make(chan reconcile.Request)
-	fn := reconcile.Func(func(req reconcile.Request) (reconcile.Result, error) {
+	fn := reconcile.Func(func(_ context.Context, req reconcile.Request) (reconcile.Result, error) {
 		defer GinkgoRecover()
 		if !strings.HasSuffix(req.Name, nameSuffix) {
 			// From different test, ignore this request.  Etcd is shared across tests.
@@ -413,21 +533,23 @@ func doReconcileTest(nameSuffix string, stop chan struct{}, blder *Builder, mgr
 		return reconcile.Result{}, nil
 	})
 
-	if complete {
-		err := blder.Complete(fn)
-		Expect(err).NotTo(HaveOccurred())
-	} else {
-		var err error
-		var c controller.Controller
-		c, err = blder.Build(fn)
-		Expect(err).NotTo(HaveOccurred())
-		Expect(c).NotTo(BeNil())
+	for _, blder := range blders {
+		if complete {
+			err := blder.Complete(fn)
+			Expect(err).NotTo(HaveOccurred())
+		} else {
+			var err error
+			var c controller.Controller
+			c, err = blder.Build(fn)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(c).NotTo(BeNil())
+		}
 	}
 
 	By("Starting the application")
 	go func() {
 		defer GinkgoRecover()
-		Expect(mgr.Start(stop)).NotTo(HaveOccurred())
+		Expect(mgr.Start(ctx)).NotTo(HaveOccurred())
 		By("Stopping the application")
 	}()
 
@@ -491,12 +613,14 @@ func doReconcileTest(nameSuffix string, stop chan struct{}, blder *Builder, mgr
 	By("Waiting for the ReplicaSet Reconcile")
 	Eventually(ch).Should(Receive(Equal(reconcile.Request{
 		NamespacedName: types.NamespacedName{Namespace: "default", Name: deployName}})))
-
 }
 
 var _ runtime.Object = &fakeType{}
 
-type fakeType struct{}
+type fakeType struct {
+	metav1.TypeMeta
+	metav1.ObjectMeta
+}
 
 func (*fakeType) GetObjectKind() schema.ObjectKind { return nil }
 func (*fakeType) DeepCopyObject() runtime.Object   { return nil }
